from __future__ import annotations

import socket
import subprocess
from threading import Thread


def _accept_and_close(sock: socket.socket) -> None:
    sock.listen()
    conn, addr = sock.accept()
    conn.close()


def check_docker_to_host_connectivity(timeout: int = 3) -> bool:
    sock = socket.socket()
    sock.bind(("127.0.0.1", 0))
    port = sock.getsockname()[1]

    listener = Thread(target=_accept_and_close, args=(sock,))
    listener.start()

    rc = subprocess.call(
        (
            "docker",
            "run",
            "--rm",
            "--add-host=host.docker.internal:host-gateway",
            "busybox:1.36.1-musl",
            "/bin/sh",
            "-c",
            f"/bin/echo hi | /bin/nc -w {timeout} host.docker.internal {port}",
        )
    )

    if rc != 0:
        # easiest way to terminate the socket server
        # (so the thread doesn't indefinitely hang)
        with socket.socket() as s:
            s.connect(("127.0.0.1", port))
            s.send(b"die")

        listener.join()
        return False

    listener.join()
    return True



from __future__ import annotations

import os
import platform
import shutil
import tempfile
from enum import Enum
from typing import Optional

from devenv.constants import home
from devenv.constants import root
from devenv.lib import archive
from devenv.lib import docker
from devenv.lib import fs
from devenv.lib import proc
from devenv.lib import rosetta


ColimaStatus = Enum("ColimaStatus", ("UP", "DOWN", "UNHEALTHY"))


def _install(url: str, sha256: str, into: str) -> None:
    os.makedirs(into, exist_ok=True)
    with tempfile.TemporaryDirectory(dir=into) as tmpd:
        archive.download(url, sha256, dest=f"{tmpd}/colima")
        os.replace(f"{tmpd}/colima", f"{into}/colima")
        os.chmod(f"{into}/colima", 0o775)


def uninstall(binroot: str) -> None:
    for d in (f"{home}/.lima",):
        shutil.rmtree(d, ignore_errors=True)

    for fp in (f"{binroot}/colima",):
        try:
            os.remove(fp)
        except FileNotFoundError:
            # it's better to do this than to guard with
            # os.path.exists(fp) because if it's an invalid or circular
            # symlink the result'll be False!
            pass


def install(
    version: str, url: str, sha256: str, reporoot: Optional[str] = ""
) -> None:
    if reporoot:
        binroot = fs.ensure_binroot(reporoot)
    else:
        # compatibility with devenv <= 1.4.0
        binroot = f"{root}/bin"
        os.makedirs(binroot, exist_ok=True)

    if shutil.which("colima", path=binroot) == f"{binroot}/colima":
        stdout = proc.run((f"{binroot}/colima", "--version"), stdout=True)
        installed_version = stdout.strip().split()[-1]
        if version == installed_version:
            return
        print(f"installed colima {installed_version} is outdated!")

    print(f"installing colima {version}...")
    uninstall(binroot)
    _install(url, sha256, binroot)

    stdout = proc.run((f"{binroot}/colima", "--version"), stdout=True)
    if f"colima version {version}" not in stdout:
        raise SystemExit(f"Failed to install colima {version}! Found: {stdout}")


def check(reporoot: str) -> ColimaStatus:
    if not os.getenv("CI"):
        macos_version = platform.mac_ver()[0]
        macos_major_version = int(macos_version.split(".")[0])
        if macos_major_version < 14:
            raise SystemExit(
                f"macos >= 14 is required to use colima, found {macos_version}"
            )

    docker_executable = shutil.which("docker")
    if not docker_executable:
        raise SystemExit(
            "docker executable not found, you might want to run devenv sync"
        )

    colima = f"{reporoot}/.devenv/bin/colima"
    if not os.path.isfile(colima):
        raise SystemExit(
            f"colima not found at {colima}, you might want to run devenv sync"
        )

    if not os.path.exists(f"{home}/.colima/default/docker.sock"):
        return ColimaStatus.DOWN

    # if colima's up, we should be able to communicate with that docker socket
    # at the most basic level
    try:
        proc.run(
            (docker_executable, "--context=colima", "version"), stdout=True
        )
    except RuntimeError:
        return ColimaStatus.UNHEALTHY

    # https://github.com/abiosoft/colima/issues/949
    healthy = docker.check_docker_to_host_connectivity()
    if not healthy:
        return ColimaStatus.UNHEALTHY

    return ColimaStatus.UP


def start(reporoot: str, restart: bool = False) -> ColimaStatus:
    status = check(reporoot)

    if status == ColimaStatus.UP:
        if not restart:
            return ColimaStatus.UP
        proc.run(("colima", "stop"), pathprepend=f"{reporoot}/.devenv/bin")
    elif status == ColimaStatus.DOWN:
        pass
    elif status == ColimaStatus.UNHEALTHY:
        print("colima seems to be unhealthy, stopping it")
        proc.run(("colima", "stop"), pathprepend=f"{reporoot}/.devenv/bin")

    # colima start will only WARN if rosetta is unavailable and keep going without it,
    # so we need to ensure it's installed and running ourselves
    rosetta.ensure()

    cpus = os.cpu_count()
    if cpus is None:
        raise SystemExit("failed to determine cpu count")

    # SC_PAGE_SIZE is POSIX 2008
    # SC_PHYS_PAGES is a linux addition but also supported by more recent MacOS versions
    SC_PAGE_SIZE = os.sysconf("SC_PAGE_SIZE")
    SC_PHYS_PAGES = os.sysconf("SC_PHYS_PAGES")
    if SC_PAGE_SIZE == -1 or SC_PHYS_PAGES == -1:
        raise SystemExit("failed to determine memsize_bytes")
    memsize_bytes = os.sysconf("SC_PAGE_SIZE") * os.sysconf("SC_PHYS_PAGES")

    args = ["--cpu", f"{cpus//2}", "--memory", f"{memsize_bytes//(2*1024**3)}"]
    if platform.machine() == "arm64":
        args = [*args, "--vm-type=vz", "--vz-rosetta", "--mount-type=virtiofs"]

    proc.run(
        (
            # we share the "default" machine across repositories
            "colima",
            "start",
            "--verbose",
            # ideally we keep ~ ro and reporoot rw, but currently the "default" vm
            # is shared across repositories, so for ease of use we'll let home rw
            f"--mount=/var/folders:w,/private/tmp/colima:w,{home}:w",
            *args,
        ),
        pathprepend=f"{reporoot}/.devenv/bin",
    )

    proc.run(("docker", "context", "use", "colima"))

    status = check(reporoot)
    return status


def restart(reporoot: str) -> ColimaStatus:
    status = start(reporoot, restart=True)
    return status


def stop(reporoot: str) -> ColimaStatus:
    proc.run(("colima", "stop"), pathprepend=f"{reporoot}/.devenv/bin")
    status = check(reporoot)
    return status



from __future__ import annotations

import os

from devenv.constants import DARWIN
from devenv.constants import MACHINE
from devenv.lib import proc


def ensure() -> None:
    if not (DARWIN and (MACHINE == "arm64")):
        return

    # this file doesn't exist if rosetta 2 isn't installed
    # (alternatively: try arch -x86_64 ... and it should fail with "bad CPU type")
    if not os.path.exists("/Library/Apple/usr/libexec/oah/libRosettaRuntime"):
        proc.run(
            ("softwareupdate", "--install-rosetta", "--agree-to-license"),
            exit=True,
        )

    # if rosetta 2 is installed and not running, the fastest way to start
    # /usr/libexec/rosetta/oahd is to request an emulated program with /usr/bin/arch
    stdout = proc.run(
        ("/usr/bin/arch", "-x86_64", "/bin/sh", "-c", "/usr/bin/uname -m"),
        stdout=True,
    )

    # we can also verify it's actually working, compared to pgrep -f /usr/libexec/rosetta/oahd
    # this is also more precise than merely checking sysctl -n sysctl.proc_translated == 1
    if stdout.strip() != "x86_64":
        raise SystemExit(f"Rosetta 2 not working as expected.\ndebug: {stdout}")



from __future__ import annotations

import argparse
import sys
from collections.abc import Sequence

from devenv import constants
from devenv.lib import proc
from devenv.lib.context import Context
from devenv.lib.modules import DevModuleInfo

module_help = "Updates global devenv."


def main(context: Context, argv: Sequence[str] | None = None) -> int:
    if not sys.executable.startswith(f"{constants.root}/venv/bin/python"):
        raise SystemExit(
            f"""
update is only applicable to the global devenv.
This one is using: {sys.executable}

To update the global devenv, run `{constants.root}/bin/devenv update`.

To update a repo-local devenv, run `devenv sync`.

(It's the responsibility of devenv/sync.py and devenv/config.ini to
maintain a virtualenv with a devenv version for that particular repository.)
"""
        )

    parser = argparse.ArgumentParser()
    parser.add_argument("version", type=str, nargs="?")

    args = parser.parse_args(argv)

    if args.version is None:
        proc.run(
            (
                f"{constants.root}/venv/bin/pip",
                "install",
                "-U",
                "sentry-devenv",
            ),
            env={
                # better than cli flag (who knows if it'll break in the future)
                "PIP_DISABLE_PIP_VERSION_CHECK": "1"
            },
        )
    else:
        proc.run(
            (
                f"{constants.root}/venv/bin/pip",
                "install",
                f"sentry-devenv=={args.version}",
            ),
            env={
                # better than cli flag (who knows if it'll break in the future)
                "PIP_DISABLE_PIP_VERSION_CHECK": "1"
            },
        )

    return 0


module_info = DevModuleInfo(
    action=main, name=__name__, command="update", help=module_help
)



from __future__ import annotations

import argparse
import os
from collections.abc import Sequence

from devenv import bootstrap
from devenv import colima
from devenv import doctor
from devenv import fetch
from devenv import pin_gha
from devenv import sync
from devenv import update
from devenv.constants import home
from devenv.constants import user
from devenv.constants import version
from devenv.lib.config import read_config
from devenv.lib.context import Context
from devenv.lib.fs import gitroot
from devenv.lib.modules import DevModuleInfo
from devenv.lib.modules import ExitCode
from devenv.lib.repository import Repository


def devenv(argv: Sequence[str], config_path: str) -> ExitCode:
    # determine current repo, if applicable
    fake_reporoot = os.getenv("CI_DEVENV_INTEGRATION_FAKE_REPOROOT")
    if fake_reporoot:
        current_root = fake_reporoot
    else:
        try:
            current_root = gitroot()
        except RuntimeError:
            current_root = None

    # This may or may not exist
    config = read_config(config_path)

    # Guessing temporary code root
    code_root = config.get("devenv", "coderoot", fallback=None) or (
        os.path.abspath(f"{current_root}/..")
        if current_root
        else os.path.expanduser("~/code")
    )

    modinfo_list: Sequence[DevModuleInfo] = [
        module.module_info
        for module in [bootstrap, fetch, colima, doctor, pin_gha, sync, update]
        if hasattr(module, "module_info")
    ]

    # TODO: Search for modules in work repo

    parser = argparse.ArgumentParser()
    parser.add_argument("--version", action="version", version=version)
    subparser = parser.add_subparsers(
        title=argparse.SUPPRESS,
        metavar="command",
        dest="command",
        required=True,
    )

    for info in modinfo_list:
        # Argparse stuff
        subparser.add_parser(info.command, help=info.help)

    if len(argv) == 1:
        parser.print_help()
        return 0

    args, remainder = parser.parse_known_args(argv[1:])

    # context for subcommands
    context: Context = {
        "config_path": config_path,
        "code_root": code_root,
        "repo": Repository(current_root) if current_root else None,
    }

    command_actions = {info.command: info.action for info in modinfo_list}
    action = command_actions.get(args.command)
    assert action is not None
    return action(context, remainder)


def main() -> ExitCode:
    if user == "root":
        raise SystemExit("You shouldn't be running devenv as root.")

    import sys
    import sentry_sdk

    sentry_sdk.init(
        # https://sentry.sentry.io/settings/projects/sentry-dev-env/keys/
        dsn="https://9bdb053cb8274ea69231834d1edeec4c@o1.ingest.sentry.io/5723503",
        # enable performance monitoring
        enable_tracing=True,
    )

    return devenv(sys.argv, f"{home}/.config/sentry-devenv/config.ini")


if __name__ == "__main__":
    raise SystemExit(main())



from __future__ import annotations

import argparse
from collections.abc import Sequence

from devenv.lib import colima
from devenv.lib.context import Context
from devenv.lib.modules import DevModuleInfo
from devenv.lib.modules import require_repo

module_help = "Colima convenience commands."


@require_repo
def main(context: Context, argv: Sequence[str] | None = None) -> int:
    repo_path = context["repo"].path  # type: ignore

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "command", choices=("start", "restart", "check", "stop")
    )

    args = parser.parse_args(argv)

    if args.command == "start":
        status = colima.start(repo_path)
        if status == colima.ColimaStatus.UNHEALTHY:
            # https://github.com/abiosoft/colima/issues/949
            print("colima seems unhealthy, we'll try restarting once")
            status = colima.restart(repo_path)
            if status != colima.ColimaStatus.UP:
                return 1
        if status != colima.ColimaStatus.UP:
            return 1
    elif args.command == "restart":
        status = colima.restart(repo_path)
        if status != colima.ColimaStatus.UP:
            return 1
    elif args.command == "check":
        status = colima.check(repo_path)
        if status != colima.ColimaStatus.UP:
            return 1
    elif args.command == "stop":
        status = colima.stop(repo_path)
        if status != colima.ColimaStatus.DOWN:
            return 1

    return 0


module_info = DevModuleInfo(
    action=main, name=__name__, command="colima", help=module_help
)
